from ...mcp_manager.client import ClientManagerMCP


class LLM:
    def __init__(self):
        self.client_manager_mcp: ClientManagerMCP = ClientManagerMCP()
        self.current_language: str = "English"

    def preprocess_query(self, query: str) -> dict:
        """
        Devuelve una lista con todas las querys en las que se divide la query principal y el lenguaje usado en la query. Puede ser solo una query
        """
        Exception("Not Implemented")

    def append_chat_history(self):
        Exception("Not Implemented Exception")

    def simple_query(self, query: str, use_services_contex: bool = False) -> str:
        Exception("Not Implemented Exception")

    def select_service(self, query: str) -> str:
        Exception("Not Implemented Exception")

    def generate_args(self, query: str, service: str) -> str:
        Exception("Not Implemented Exception")

    def final_response(self, query: str, data: str | dict) -> str:
        Exception("Not Implemented Exception")


"""Call an Auth Admin method from Supabase Python SDK.\n\nThis tool provides a safe, validated interface to the Supabase Auth Admin SDK, allowing you to:\n- Manage users (create, update, delete)\n- List and search users\n- Generate authentication links\n- Manage multi-factor authentication\n- And more\n\nIMPORTANT NOTES:\n- Request bodies must adhere to the Python SDK specification\n- Some methods may have nested parameter structures\n- The tool validates all parameters against Pydantic models\n- Extra fields not defined in the models will be rejected\n\nAVAILABLE METHODS:\n- get_user_by_id: Retrieve a user by their ID\n- list_users: List all users with pagination\n- create_user: Create a new user\n- delete_user: Delete a user by their ID\n- invite_user_by_email: Send an invite link to a user\'s email\n- generate_link: Generate an email link for various authentication purposes\n- update_user_by_id: Update user attributes by ID\n- delete_factor: Delete a factor on a user\n\nEXAMPLES:\n1. Get user by ID:\n   method: "get_user_by_id"\n   params: {"uid": "user-uuid-here"}\n\n2. Create user:\n   method: "create_user"\n   params: {\n     "email": "user@example.com",\n     "password": "secure-password"\n   }\n\n3. Update user by ID:\n   method: "update_user_by_id"\n   params: {\n     "uid": "user-uuid-here",\n     "attributes": {\n       "email": "new@email.com"\n     }\n   }\n\nFor complete documentation of all methods and their parameters, use the get_auth_admin_methods_spec tool.\n"""



from mcp.client.streamable_http import streamablehttp_client
from mcp import ClientSession
import asyncio


def call_tool(http: str, toolname: str, args: dict) -> any:
    return asyncio.run(async_call_tool(http, toolname, args))


async def async_call_tool(http: str, toolname: str, args: dict):
    async with streamablehttp_client(http) as (
        read_stream,
        write_stream,
        _,
    ):
        async with ClientSession(read_stream, write_stream) as session:
            await session.initialize()

            # Call a tool
            tool_result = await session.call_tool(toolname, args)
            return tool_result.content



from ...prompts.system_prompts import (
    chat_asistant,
    select_service,
    create_args,
    preproccess_query,
    language_prompt,
)
from ...prompts.user_prompts import query_and_data, query_and_services, service2args
from .....config.llm_config import ConfigGPT
from ...llm import LLM
from .....mcp_manager.services.service import Service
from openai import OpenAI, AsyncOpenAI
import json


class GPT(LLM):
    """
    ## GPT
    ### Args:
    - `model`: modelo de gpt que a usar
    - `max_history_len`: largo maximo del historial que se para como input al modelo
    """

    def __init__(
        self,
        model=ConfigGPT.DEFAULT_MODEL_NAME,
        max_history_len: int = 10,
    ):
        super().__init__()
        self.client = OpenAI(api_key=ConfigGPT.OPENAI_API_KEY)
        """Cliente secuencial de GPT"""
        self.asyncclient = AsyncOpenAI(api_key=ConfigGPT.OPENAI_API_KEY)
        """Cliente asincrono de GPT"""
        self.max_len_history: int = max_history_len
        """Maxima cantidad de mensajes previos que se le pasan como input"""
        self.chat_history: list[list[dict[str, str]]] = []
        """Historial del chat asosiado a esta instancia de GPT, en forma lista de listas, por ejemplo 
        ```
        chat_history: str = [
            [
                {"role": "user", "content": "query1"},
                {"role": "assistant", "content": "response1"}
            ],
            [
                {"role": "user", "content": "query2"},
                {"role": "assistant", "content": "response2"}
            ]
        ]
        ```
        """
        self.model: str = model
        self.current_price: float = 0

    def append_chat_history(self):
        """Agrega una historia vacia para crear el espacio"""
        self.chat_history.append([])
        self.chat_history[-1].append({})

    def call_completion(
        self, system_message: str, query: str, json_format: bool = False
    ):
        """ """
        self.chat_history[-1][0] = {"role": "user", "content": query}

        messages: list[dict[str, str]] = [
            {"role": "system", "content": system_message}
        ] + [
            message
            for messages in self.chat_history[-self.max_len_history :]
            for message in messages
        ]
        if json_format:
            completion = self.client.chat.completions.create(
                model=self.model,
                messages=messages,
                response_format={"type": "json_object"},
            )
        else:

            completion = self.client.chat.completions.create(
                model=self.model,
                messages=messages,
            )
        self.get_price(completion.usage)
        return completion.choices[0].message.content

    def get_price(self, usage) -> float:
        """
        ### Args
            - `usage`: uso de la api retornado en el completion respuesta del llamado a la api de gpt
        ### Outs:
            - `price`: precio final del llamado a la api.
        """

        input_tokens: int = usage.prompt_tokens
        output_tokens: int = usage.completion_tokens

        input_price: float = ConfigGPT.MODEL_PRICE[self.model]["input"]
        output_price: float = ConfigGPT.MODEL_PRICE[self.model]["output"]
        price: float = input_tokens * input_price + output_tokens * output_price

        # Aumenta el valod asociado al costo de uso de la API en esta instancia de GPT
        self.current_price += price
        return price

    def preprocess_query(self, query: str) -> dict:
        system_message: str = preproccess_query(
            services=self.client_manager_mcp.get_services()
        )
        messages: list[dict[str, str]] = [
            {"role": "system", "content": system_message},
            {"role": "user", "content": query},
        ]
        completion = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            response_format={"type": "json_object"},
        )
        self.get_price(completion.usage)
        response = completion.choices[0].message.content
        return json.loads(response)

    def simple_query(self, query: str, use_services_contex: bool = False) -> str:
        system_message: str = chat_asistant(
            self.client_manager_mcp.get_services() if use_services_contex else None
        ) + language_prompt(self.current_language)
        response = self.call_completion(system_message=system_message, query=query)

        # Se agrega la respuesta a la historia
        self.chat_history[-1].append({"role": "assistant", "content": response})
        return response

    def select_service(self, query: str) -> str:
        """
        Funcion encargada de seleccionar los servicios utiles para el contexto de la consulta, usando los servicios expuestos por cada uno
        de los servidores
        """
        system_message: str = select_service
        return self.call_completion(
            system_message=system_message,
            query=query_and_services(
                query=query, services=self.client_manager_mcp.get_services()
            ),
            json_format=True,
        )

    def generate_args(self, query: str, service: Service) -> str:
        """
        Funcion encargada de crear argumentos para los servicios expuestos que se usaran
        """
        system_message: str = create_args

        service = str(service)

        return self.call_completion(
            system_message=system_message,
            query=service2args(query=query, service=service),
            json_format=True,
        )

    def final_response(self, query: str, data: str | dict) -> str:
        system_message: str = chat_asistant() + language_prompt(self.current_language)
        user_message = query_and_data(query=query, data=data)

        response = self.call_completion(
            system_message=system_message,
            query=user_message,
            json_format=False,
        )

        # Se agrega la respuesta a la historia
        self.chat_history[-1].append({"role": "assistant", "content": response})
        return response



def language_prompt(language: str) -> str:
    return f"\nAlways respond in the language: {language}"


def chat_asistant(services: list | None = None) -> str:
    return (
        "You are an advanced assistant with the ability to maintain professional and friendly conversations. Your primary function is to provide detailed and structured responses, adapting to the specific context of each query while maintaining an accessible and professional tone. As a customer service expert, you specialize in completely understanding the context of each interaction, providing structured and detailed responses, maintaining a coherent and consistent dialogue, adapting to the technical level and style of the user, and offering precise and useful solutions."
        + (
            ""
            if services is None
            else f"\n You have access to the following aviable services:\n{services}"
        )
    )


select_service: str = (
    """You are an expert in data comprehension and have access to several services. Given the user's query and the services that pass you with the format JSON {key1 : service_data1, key2 : service_data2, ...}, your task is to define whether there exists or not a useful service for the query context, in case there exists extract the service that will be useful for the given context; in case there is no useful service for the context you must return empty value {"service":""}. Ensure returning a JSON with format {"service":"key of the service selected for the context"}. Ensure selecting services correctly, an empty service ("") is a valid selection."""
)

create_args: str = (
    """You are an expert in data comprehension and have access to various useful services to respond to the context. Given the user's query and a service that will be passed to you in the same JSON format {"name":"service name", "description":"service description","args": "input arguments of service"}, your task is to conform the list of arguments necessary for the service and return it in the JSON format: `{"args":{"arg1 name":"arg1 value", "arg2 name":"arg2 value",...} }`"""
)


def preproccess_query(services: list) -> str:
    return (
        """You are an expert in task comprehension and ordering of the same. Your mission is, given a user's query, to separate it into several independent queries if necessary. If the query doesn't need to be separated into more than one task then you must return a list of size 1 with exactly the same user's query. It's important that you separate them in the correct order of execution according to their dependencies on each other. The condition for separating queries is given by a list of available services, if it's necessary to use more than one service then you must separate the query. Also you must extract the language used  in the query, it can be any language.\n"""
        + """For example, if the query is: "Extract the information of the user with id 222 from database 1 and add this user to database 2" and you have a service to consult the database and another to add to the database; then you must separate the query into two subqueries: {"querys":["Extract the information of the user with id 222 from database 1", "Add this user's information to database 2"],"language":"language used in the query"}.\n"""
        + 'You must return the response in a JSON format with the structure: `{"querys":[list of each of the resulting queries]}`.\n'
        + f"The list of available services is the following:\n {services}"
    ).replace("'", '"')



from .servers import Servers
from .connections.session_data import get_session_data
from .connections.call_tool import call_tool
from .connections.read_resource import read_resource
from ..tools.get_uri_args import get_args_from_uri
from .services.service import Tool, Resource, Prompt, Service
from ..config.logger import logger
from mcp import ClientSession


class ClientManagerMCP:
    def __init__(self):
        self.tools: dict[str:Tool] | None = {}
        self.resources: dict[str:Resource] | None = {}
        self.prompts: dict[str:Prompt] | None = {}
        self.refresh_data()
        self.services: list[dict] = []
        """Lista de servicios en forma de string para pasarse al LLM"""

    def call_tool(self, name: str, args: dict) -> str | None:
        return self.tools[name].call()

    def read_resource(self, name: str, args: dict) -> str | None:
        return self.resources[name].read()

    def get_promps(self) -> dict:
        pass

    def refresh_data(self):
        """
        Inicializa o refresca la lista de herramientas, recursos o promps que sirve cada uno de los servidores, y lo organiza en forma de diccionario
        donde cada valor posee informacion de cada servicio ademas de la direccion desde la cual se sirve
        """
        self.tools, self.resources, self.prompts = ({}, {}, {})
        mcp_servers: dict[str, dict] = Servers().mcp_servers

        for server_key in mcp_servers.keys():
            server = {"key": server_key} | mcp_servers[server_key]
            try:
                session: dict = get_session_data(server["http"])
            except Exception as e:
                logger.warning(
                    f"Failed to establish connection with server {server_key}. Cause: {e}"
                )
                continue
            for tool in session["tools"]:
                self.tools[f"{server_key}_{tool.name}"] = Tool(
                    http=server["http"], data=tool, server=server
                )
            for resource in session["resources"]:
                self.resources[f"{server_key}_{resource.name}"] = Resource(
                    http=server["http"], data=resource, server=server
                )

            for prompt in session["prompts"]:
                self.prompts[f"{server_key}_{prompt.name}"] = Prompt(
                    http=server["http"], data=prompt, server=server
                )

    def service_type(self, service_key: str) -> str:
        if self.tools.keys().__contains__(service_key):
            return "tool"
        if self.resources.keys().__contains__(service_key):
            return "resource"

    def get_services(self):
        if len(self.services) != len(self.tools) + len(self.resources):
            services = self.tools | self.resources
            self.services = [
                {service: str(services[service])} for service in services.keys()
            ]

        return self.services



import logging
import logging.handlers
from pathlib import Path
import sys


class CustomFormatter(logging.Formatter):
    grey = "\x1b[38;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "{color}%(levelname)s\x1b[0m: %(asctime)s  -  %(message)s"

    FORMATS = {
        logging.DEBUG: format.replace("{color}", grey),
        logging.INFO: format.replace("{color}", yellow),
        logging.WARNING: format.replace("{color}", yellow),
        logging.ERROR: format.replace("{color}", red),
        logging.CRITICAL: format.replace("{color}", bold_red),
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def setup_logger() -> logging.Logger:
    """Configure logging for the MCP client with log rotation."""
    logger = logging.getLogger("client-mcp")

    # Remove existing handlers to avoid duplicate logs
    if logger.hasHandlers():
        logger.handlers.clear()

    # Define a consistent log directory in the user's home folder
    log_dir = Path.home() / ".local" / "share" / "mcp-llm-client"
    log_dir.mkdir(parents=True, exist_ok=True)  # Ensure the directory exists

    # Define the log file path
    log_file = log_dir / "mcp_client.log"

    # Create a rotating file handler
    # - Rotate when log reaches 5MB
    # - Keep 3 backup files
    file_handler = logging.handlers.RotatingFileHandler(
        log_file,
        maxBytes=5 * 1024 * 1024,  # 5MB
        backupCount=3,
        encoding="utf-8",
    )

    # Add formatter to file handler
    file_handler.setFormatter(CustomFormatter())

    # Add handler to logger
    logger.addHandler(file_handler)

    # Terminal Handler  (stdout)
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(CustomFormatter())
    logger.addHandler(console_handler)

    # Set level
    logger.setLevel(logging.DEBUG)

    return logger


logger = setup_logger()



from mcp.client.streamable_http import streamablehttp_client
from mcp import ClientSession
import asyncio


def get_session_data(http: str) -> dict:
    return asyncio.run(async_get_session(http))


async def async_get_session(http: str):
    # Connect to a streamable HTTP server
    async with streamablehttp_client(http) as (
        read_stream,
        write_stream,
        _,
    ):
        # Create a session using the client streams
        async with ClientSession(read_stream, write_stream) as session:
            # Initialize the connection
            await session.initialize()
            
            # Generate data 
            tools = await session.list_tools()
            resources = await session.list_resource_templates()
            prompts = await session.list_prompts()

            data: dict = {
                "tools": tools.tools,
                "resources": resources.resourceTemplates,
                "prompts": prompts.prompts,
            }
            

        return data



def service2args(query: str, service: dict):
    return "{" + f'"User Query": {query}, "Service":{service}' + "}"
    # return f'{query}.\n Use the following service as information base:\n{service}'


def query_and_services(query: str, services: str | list):
    return "{" + f'"User Query": {query}, "All Services":{services}' + "}"
    # return f"{query}.\n The available services for obtaining information are:\n{services}"


def query_and_data(query: str, data: str | dict):
    return f"{query}.\n Use the following retrievd data to answer my query:\n{data}"
    # return "{" + f'"User Query": {query}, "Context":{data}' + "}"



import os
import json


class Servers:
    """
    Extrae las credenciales de cada uno de los servers en la configuracion `.config.json`. Aternativamente puede pasarse otro archivo `.json`
    como argumento `config_file_path`
    """

    def __init__(self, config_file_path: str = "config.json"):
        self.config_file_path: str = config_file_path
        self.mcp_servers: dict[str, dict] | None = {}
        self.load_servers()

    def load_servers(self):
        """Carga cada uno de los credenciales de los servidores que se le conectan desde el `.config`"""
        with open(f"{os.getcwd()}{os.path.sep}{self.config_file_path}", "r") as file:
            json_config: dict = json.loads(file.read())
        self.mcp_servers = json_config["mcp_servers"]



from ...tools.get_uri_args import get_args_from_uri
from mcp.client.streamable_http import streamablehttp_client
from mcp import ClientSession
import asyncio


class Service:
    """
    Objeto que representa un servicio brindado o expuesto por un servidor mcp. Falicita la manipulacion del mismo
    """

    def __init__(self, http: str, data, server: dict):
        self.http: str = http
        self.data = data
        self.name: str = data.name
        self.description: str = data.description
        self.args: dict[str, any] = None
        self.server: dict = server

    def __str__(self):
        return str(
            {
                "name": self.name,
                "description": self.description,
                "args": self.args,
            }
        )

    def __call__(self, args: dict[str, any]):
        Exception("Not Implemented")


class Tool(Service):
    def __init__(self, http, data, server):
        super().__init__(http, data, server)
        args = data.inputSchema["properties"]
        self.args = [{"name": key, "type": args[key]["type"]} for key in args.keys()]

    def __call__(self, args: dict[str, any]):
        return self.call(args)

    def call(self, args: dict[str, any]):
        return asyncio.run(Tool.async_call(self.http, self.name, args))

    async def async_call(http: str, toolname: str, args: dict):
        async with streamablehttp_client(http) as (
            read_stream,
            write_stream,
            _,
        ):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                # Call a tool
                tool_result = await session.call_tool(toolname, args)
                return tool_result.content


class Resource(Service):
    def __init__(self, http, data, server):
        super().__init__(http, data, server)
        self.args = get_args_from_uri(data.uriTemplate)
        self.args = [{"name": arg, "type": "string"} for arg in self.args]

    def __call__(self, args: dict[str, any]):
        return self.read(args)

    def read(self, args: dict[str, str]):
        uri = self.data.uriTemplate
        for key in args:
            uri = uri.replace("{" + key + "}", str(args[key]))
        return asyncio.run(Resource.async_read(self.http, uri))

    async def async_read(http: str, uri: str):
        async with streamablehttp_client(http) as (
            read_stream,
            write_stream,
            _,
        ):
            async with ClientSession(read_stream, write_stream) as session:
                await session.initialize()

                # Read a resource
                resource_result = await session.read_resource(uri)
                return resource_result.contents


class Prompt(Service):
    def __init__(self, http, data, server):
        super().__init__(http, data, server)

    def __call__(self, args: dict[str, any]):
        return self.get(args)

    def get(self, args: dict[str, str]):
        pass



from mcpclient.services.llm.chat.chat import Chat
from mcpclient.tools.clear_console import clear_console
import os


def open_local_chat():
    clear_console()
    chat: Chat = Chat()
    print("\n")

    md: str = ""
    filename: str | None = None
    sep = os.path.sep

    while True:
        query = input("> ")
        if query == "exit" or query == "":
            if md != "":
                chats_path = f"{os.getcwd()}{sep}chats{sep}"
                os.makedirs(os.path.dirname(chats_path), exist_ok=True)
                with open(f"{chats_path}{filename}.md", "w") as file:
                    file.write(md)
            break

        if filename is None:
            filename = f"{query[:15]}..."

        md += f"# {query}"
        index = 1
        for step in chat(query):
            md += str(step)
            if step.type == "response":
                print(f"<< {step.response}")
            if step.type == "query":
                print(f">> {step.query}")
                index = 1
            if step.type == "data":
                print(f'   {str(step).replace("*", "").replace("- ", "• ")}')
            if step.type == "step":
                print(f"  {index}. {step.step}")
                index += 1
        md += "\n"



import os


class ConfigLLM:
    USED_LLM: str | None = None


class ConfigGPT:
    MODEL_PRICE = {
        "gpt-3.5-turbo": {
            "input": 0.5 / 1000000,
            "output": 1.5 / 1000000,
        },
        "gpt-4o-mini": {
            "input": 0.15 / 1000000,
            "output": 0.60 / 1000000,
        },
    }
    """Price by one tokens from each model"""

    DEFAULT_MODEL_NAME = "gpt-4o-mini"
    """GPT model that will be used"""

    OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY")
    """OpenAI Api Key"""



def get_args_from_uri(uri: str) -> list[str]:
    result = []
    current_arg = ""
    for char in uri:
        if char == "{":
            current_arg = ""
        elif char == "}":
            result.append(current_arg)
        else:
            current_arg += char
            
    return result



from ..llm import LLM
from ..models.openai_service.gpt import GPT
from typing import Generator
from .step import Step
from .step import Step, StepMessage, DataStep, ResponseStep, QueryStep
import json


class Chat:
    def __init__(
        self,
        name="openai",
        model="gpt4o-mini",
        len_context: int = 10,
        history: list = [],
        id: str | None = None,
    ):
        self.id = id
        self.llm: LLM = GPT(max_history_len=len_context)

    def __call__(self, query: str) -> Generator[Step]:
        yield Step(step_type=StepMessage.ANALYZE_QUERY)

        processed_query: dict = self.llm.preprocess_query(query=query)
        querys: list[str] = processed_query["querys"]
        self.llm.current_language = processed_query["language"]

        yield DataStep(data={"querys": querys})

        for query in querys:
            for step in self.proccess_query(query):
                yield step

    def proccess_query(self, query: str) -> Generator[Step]:
        """ """
        self.llm.append_chat_history()
        yield QueryStep(query)
        yield Step(step_type=StepMessage.SELECT_SERVICE)
        # Cargar los servicios utiles
        service = json.loads(self.llm.select_service(query))["service"]

        if len(service) == 0:
            yield DataStep(data={"service": None})
            yield ResponseStep(
                response=self.llm.simple_query(query, use_services_contex=True),
                data=None,
            )
            return
        else:
            yield DataStep(data={"service": service})
            service = (
                self.llm.client_manager_mcp.resources[service]
                if (self.llm.client_manager_mcp.service_type(service) == "resource")
                else (
                    self.llm.client_manager_mcp.tools[service]
                    if (self.llm.client_manager_mcp.service_type(service) == "tool")
                    else None
                )
            )

        yield Step(step_type=StepMessage.CREATE_ARGUMENTS)
        args = json.loads(self.llm.generate_args(query=query, service=service))["args"]
        data = service(args)[0].text
        yield DataStep(data={"args": args})

        response = self.llm.final_response(query, data)
        yield ResponseStep(response=response, data=data)


from mcp.client.streamable_http import streamablehttp_client
from mcp import ClientSession
import asyncio


def read_resource(http: str, uri: str) -> any:
    return asyncio.run(async_read_resource(http, uri))


async def async_read_resource(http: str, uri: str):
    async with streamablehttp_client(http) as (
        read_stream,
        write_stream,
        _,
    ):
        async with ClientSession(read_stream, write_stream) as session:
            await session.initialize()

            # Read a resource
            resource_result = await session.read_resource(uri)
            return resource_result.contents



from enum import Enum


class StepMessage(Enum):
    """
    Enum representing all available LLM processing steps.
    Each step includes a name and a clear description of its purpose.
    """

    ANALYZE_QUERY = (
        "analyze query",
        "Analyze the user query and available services to separate it into service-specific calls.",
    )

    SELECT_SERVICE = (
        "select service",
        "Select the most appropriate service for the current query.",
    )

    CREATE_ARGUMENTS = (
        "create arguments",
        "Generate the necessary arguments for use with the selected service.",
    )


class Step:
    def __init__(self, step_type: StepMessage, data: dict[str, any] | None = None):
        self.type: str = "step"
        self.step: str = step_type.value[0]
        self.message: str = step_type.value[1]
        self.data: dict[str, any] = data

        self.json: dict[str, any] = {
            "type": self.type,
            "step": self.step,
            "message": self.message,
            "data": self.data,
        }
        self.response: str | None = None
        self.query: str | None = None

    def __call__(self) -> dict:
        return self.json

    def __str__(self):
        result = f"\n### {self.step}\n{self.message}\n"
        if self.data is not None and len(self.data) > 0:
            # result += "#### Data\n"
            for key in self.data.keys():
                result += f"- **{key}:** {self.data[key]}\n"
        return result


class ResponseStep(Step):
    def __init__(self, response: str, data: dict[str, any] | None = None):
        self.type: str = "response"
        self.response: str | None = response
        self.data: dict[str, any] | None = data
        self.json: dict[str, any] = {
            "type": self.type,
            "response": self.response,
            "data": data,
        }

    def __str__(self):
        return f"\n### Response\n{self.response}" + (
            f"\n```json\n{self.data}\n```" if self.data is not None else ""
        )


class QueryStep(Step):
    def __init__(self, query: str):
        self.type: str = "query"
        self.query: str | None = query
        self.json: dict[str, any] = {"type": self.type, "query": self.query}

    def __str__(self):
        return f"\n## {self.query}"


class DataStep(Step):
    def __init__(self, data: dict[str, any]):
        self.type: str = "data"
        self.data: dict[str, any] | None = data
        self.json: dict[str, any] = {"type": self.type, "data": self.data}

    def __str__(self):
        result = ""
        if self.data is not None and len(self.data) > 0:
            # result += "#### Data\n"
            for key in self.data.keys():
                result += f"- **{key}:** {self.data[key]}\n"
        return result
